"""
This module defines the documented objects classes:

- the generic [`Object`][pytkdocs.objects.Object] class
- the [`Module`][pytkdocs.objects.Module] class
- the [`Class`][pytkdocs.objects.Class] class
- the [`Method`][pytkdocs.objects.Method] class
- the [`Function`][pytkdocs.objects.Function] class
- the [`Attribute`][pytkdocs.objects.Attribute] class

Note that properties are considered attributes, because they are used like such.

It also defines a convenient [`Source`][pytkdocs.objects.Source] class to represent source code.
"""

import importlib
import inspect
import os
import sys
from abc import ABCMeta
from functools import lru_cache
from pathlib import Path
from typing import List, Optional, Union

from pytkdocs.parsers.docstrings import parse

from .properties import NAME_CLASS_PRIVATE, NAME_PRIVATE, NAME_SPECIAL, ApplicableNameProperty


class Source:
    """
    Helper class to represent source code.

    It is simply used to wrap the result of
    [`inspect.getsourceslines`](https://docs.python.org/3/library/inspect.html#inspect.getsourcelines).
    """

    def __init__(self, lines: Union[str, List[str]], line_start: int) -> None:
        """
        Arguments:
            lines: A list of strings. The strings should have trailing newlines.
            line_start: The line number of where the code starts in the file.
        """
        if isinstance(lines, list):
            code = "".join(lines)
        else:
            code = lines
        self.code = code
        """The code, as a single string."""
        self.line_start = line_start
        """The first line number."""


class Object(metaclass=ABCMeta):
    """
    A base class to store information about a Python object.

    Each instance additionally stores references to its children, grouped by category.
    """

    NAME_PROPERTIES: List[ApplicableNameProperty] = []
    """
    The properties that we can apply to the object based on its name.

    The applicable properties vary from one subclass of `Object` to another.
    """

    def __init__(
        self,
        name: str,
        path: str,
        file_path: str,
        docstring: str = "",
        properties: Optional[List[str]] = None,
        source: Optional[Source] = None,
    ) -> None:
        """
        Arguments:
            name: The object's name.
            path: The object's dotted-path.
            file_path: The file path of the object's direct parent module.
            docstring: The object's docstring.
            properties: The object's properties.
            source: The object's source code.
        """

        self.name = name
        """The object's name."""
        self.path = path
        """The object's dotted-path."""
        self.file_path = file_path
        """The file path of the object's direct parent module."""
        self.docstring = docstring
        """The object's docstring."""
        self.properties = properties or []
        """The object's properties."""
        self.parent: Optional[Object] = None
        """The object's parent (another instance of a subclass of `Object`)."""
        self.source = source
        """The object's source code."""

        self._path_map = {self.path: self}

        self.attributes: List[Attribute] = []
        """The list of all the object's attributes."""
        self.methods: List[Method] = []
        """The list of all the object's methods."""
        self.functions: List[Function] = []
        """The list of all the object's functions."""
        self.modules: List[Module] = []
        """The list of all the object's submodules."""
        self.classes: List[Class] = []
        """The list of all the object's classes."""
        self.children: List[Object] = []
        """The list of all the object's children."""

    def __str__(self) -> str:
        return self.path

    @property
    def is_module(self) -> bool:
        """Is this object a module?"""
        return isinstance(self, Module)

    @property
    def is_class(self) -> bool:
        """Is this object a class?"""
        return isinstance(self, Class)

    @property
    def is_function(self) -> bool:
        """Is this object a function?"""
        return isinstance(self, Function)

    @property
    def is_method(self) -> bool:
        """Is this object a method?"""
        return isinstance(self, Method)

    @property
    def is_attribute(self) -> bool:
        """Is this object an attribute?"""
        return isinstance(self, Attribute)

    @property
    def category(self) -> str:
        """The object's category: module, class, function, method or attribute."""
        return self.__class__.__name__.lower()

    @property
    def root(self) -> "Object":
        """The object's root (top-most parent)."""
        obj = self
        while obj.parent:
            obj = obj.parent
        return obj  # type: ignore

    @property
    def relative_file_path(self) -> str:
        """
        The relative file path of the object.

        It is the relative path to the object's module,
        starting at the path of the top-most package it is contained in.

        For example:

        - package is `a`
        - package absolute path is `/abs/path/to/a`
        - module is `a.b.c`
        - object is `c` or anything defined in `c`
        - relative file path is `a/b/c.py`

        If the relative file path cannot be determined, the value returned is `""` (empty string).
        """
        top_package_name = self.path.split(".", 1)[0]
        try:
            top_package = sys.modules[top_package_name]
        except KeyError:
            try:
                importlib.import_module(top_package_name)
            except ImportError:
                return ""
            top_package = sys.modules[top_package_name]
        top_package_path = Path(inspect.getabsfile(top_package)).parent
        try:
            return str(Path(self.file_path).relative_to(top_package_path.parent))
        except ValueError:
            return ""

    @property
    def name_to_check(self) -> str:
        """The attribute to check against name properties regular expressions (private, class-private, special)."""
        return self.name

    @property
    def name_properties(self) -> List[str]:
        """The object's name properties (private, class-private, special)."""
        properties = []
        for prop, predicate in self.NAME_PROPERTIES:
            if predicate(self.name_to_check):
                properties.append(prop)
        return properties

    @property
    def parent_path(self) -> str:
        """
        The parent's path, computed from the current path.

        The parent object path is not used: this property is used to see if an object is really related to another one,
        to add it as a child to the other. When we do that, the child doesn't even have a parent.
        """
        return self.path.rsplit(".", 1)[0]

    def add_child(self, obj: "Object") -> None:
        """
        Add an object as a child of this object.

        If the child computed `parent_path` is not equal to this object's path, abort.

        Append the child to the `children` list, and to the right category list.

        Arguments:
            obj: An instance of documented object.
        """
        if obj.parent_path != self.path:
            return

        self.children.append(obj)
        if obj.is_module:
            self.modules.append(obj)  # type: ignore
        elif obj.is_class:
            self.classes.append(obj)  # type: ignore
        elif obj.is_function:
            self.functions.append(obj)  # type: ignore
        elif obj.is_method:
            self.methods.append(obj)  # type: ignore
        elif obj.is_attribute:
            self.attributes.append(obj)  # type: ignore
        obj.parent = self

        self._path_map[obj.path] = obj

    def add_children(self, children: List["Object"]) -> None:
        """
        Add a list of objects as children of this object.

        Arguments:
            children: The list of children to add.
        """
        for child in children:
            self.add_child(child)

    def dispatch_attributes(self, attributes: List["Attribute"]) -> None:
        """
        Dispatch attributes as children of an object and its children given their path.

        If an attribute's path does not correspond to the object or any of its children,
        the attribute is not attached.

        Arguments:
            attributes: The list of attributes to dispatch.
        """
        for attribute in attributes:
            try:
                attach_to = self._path_map[attribute.parent_path]
            except KeyError:
                pass
            else:
                attach_to.attributes.append(attribute)
                attach_to.children.append(attribute)
                attribute.parent = attach_to

    def parse_all_docstring(self) -> None:
        """
        Recursively parse the docstring of this object and its children.

        I hope we can get rid of this code at some point as parsing docstring is not really our purpose.
        """
        signature = None
        if hasattr(self, "signature"):
            signature = self.signature  # type: ignore
        attr_type = None
        if hasattr(self, "type"):
            attr_type = self.type  # type: ignore
        sections, errors = parse(self.path, self.docstring, signature, attr_type)
        self.docstring_sections = sections
        self.docstring_errors = errors
        for child in self.children:
            child.parse_all_docstring()

    @lru_cache()
    def has_contents(self) -> bool:
        """
        Tells if the object has "contents".

        An object has contents when:

        - it is the root of the object tree
        - it has a docstring
        - at least one of its children (whatever the depth) has contents

        The value is cached, so this method should be called last, when the tree doesn't change anymore.
        """
        return bool(self.docstring or not self.parent or any(c.has_contents() for c in self.children))


class Module(Object):
    """A class to store information about a module."""

    NAME_PROPERTIES: List[ApplicableNameProperty] = [NAME_SPECIAL, NAME_PRIVATE]

    @property
    def file_name(self):
        """The base name of the module file, without the extension."""
        return os.path.splitext(os.path.basename(self.file_path))[0]

    @property
    def name_to_check(self):
        return self.file_name


class Class(Object):
    """A class to store information about a class."""

    NAME_PROPERTIES: List[ApplicableNameProperty] = [NAME_PRIVATE]


class Function(Object):
    """
    A class to store information about a function.

    It accepts an additional `signature` argument at instantiation.
    """

    NAME_PROPERTIES: List[ApplicableNameProperty] = [NAME_PRIVATE]

    def __init__(self, *args, signature=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.signature = signature


class Method(Object):
    """
    A class to store information about a method.

    It accepts an additional `signature` argument at instantiation.
    """

    NAME_PROPERTIES: List[ApplicableNameProperty] = [NAME_SPECIAL, NAME_PRIVATE]

    def __init__(self, *args, signature=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.signature = signature


class Attribute(Object):
    """
    A class to store information about an attribute.

    It accepts an additional `attr_type` argument at instantiation.
    """

    NAME_PROPERTIES: List[ApplicableNameProperty] = [NAME_SPECIAL, NAME_CLASS_PRIVATE, NAME_PRIVATE]

    def __init__(self, *args, attr_type=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.type = attr_type



"""This module simply defines regular expressions and their associated predicates."""

import re
from typing import Callable, Pattern, Tuple

ApplicableNameProperty = Tuple[str, Callable[[str], bool]]

# exactly two leading underscores, exactly two trailing underscores
# since we enforce one non-underscore after the two leading underscores,
# we put the rest in an optional group
RE_SPECIAL: Pattern = re.compile(r"^__[^_]([\w_]*[^_])?__$")
"""Regular expression to match `__special__` names."""

# at least two leading underscores, at most one trailing underscore
# since we enforce one non-underscore before the last,
# we make the previous characters optional with an asterisk
RE_CLASS_PRIVATE: Pattern = re.compile(r"^__[\w_]*[^_]_?$")
"""Regular expression to match `__class_private` names."""

# at most one leading underscore, then whatever
RE_PRIVATE: Pattern = re.compile(r"^_[^_][\w_]*$")
"""Regular expression to match `_private` names."""

NAME_SPECIAL: ApplicableNameProperty = ("special", lambda n: bool(RE_SPECIAL.match(n)))
"""Applicable property: `special`."""

NAME_CLASS_PRIVATE: ApplicableNameProperty = ("class-private", lambda n: bool(RE_CLASS_PRIVATE.match(n)))
"""Applicable property: `class-private`."""

NAME_PRIVATE: ApplicableNameProperty = ("private", lambda n: bool(RE_PRIVATE.match(n)))
"""Applicable property: `private`."""



"""
This module contains the command line application.

Why does this file exist, and why not put this in `__main__`?

You might be tempted to import things from __main__ later,
but that will cause problems; the code will get executed twice:

- When you run `python -m pytkdocs` python will execute
  `__main__.py` as a script. That means there won't be any
  `pytkdocs.__main__` in `sys.modules`.
- When you import __main__ it will get executed again (as a module) because
  there's no `pytkdocs.__main__` in `sys.modules`.

Also see http://click.pocoo.org/5/setuptools/#setuptools-integration.
"""

import argparse
import json
import sys
import traceback
from typing import Dict, List, Optional, Sequence

from .loader import Loader
from .objects import Object
from .serializer import serialize_object


def process_config(config: dict) -> dict:
    """
    Process a loading configuration.

    The `config` argument is a dictionary looking like this:

    ```python
    {
        "objects": [
            {"path": "python.dotted.path.to.the.object1"},
            {"path": "python.dotted.path.to.the.object2"}
        ]
    }
    ```

    The result is a dictionary looking like this:

    ```python
    {
        "loading_errors": [
            "message1",
            "message2",
        ],
        "parsing_errors": {
            "path.to.object1": [
                "message1",
                "message2",
            ],
            "path.to.object2": [
                "message1",
                "message2",
            ]
        },
        "objects": [
            {
                "path": "path.to.object1",
                # other attributes, see the documentation for `pytkdocs.objects` or `pytkdocs.serializer`
            },
            {
                "path": "path.to.object2",
                # other attributes, see the documentation for `pytkdocs.objects` or `pytkdocs.serializer`
            },
        ]
    }
    ```

    Arguments:
        config: The configuration.

    Returns:
        The collected documentation along with the errors that occurred.
    """
    collected = []
    loading_errors = []
    parsing_errors = {}

    for obj_config in config["objects"]:
        path = obj_config.pop("path")
        filters = obj_config.get("filters", [])
        members = obj_config.get("members", set())
        if isinstance(members, list):
            members = set(members)
        loader = Loader(filters=filters)

        obj = loader.get_object_documentation(path, members)

        loading_errors.extend(loader.errors)
        parsing_errors.update(extract_errors(obj))

        serialized_obj = serialize_object(obj)
        collected.append(serialized_obj)

    return dict(loading_errors=loading_errors, parsing_errors=parsing_errors, objects=collected)


def process_json(json_input: str) -> dict:
    """
    Process JSON input.

    Simply load the JSON as a Python dictionary, then pass it to [`process_config`][pytkdocs.cli.process_config].

    Arguments:
        json_input: The JSON to load.

    Returns:
        The result of the call to [`process_config`][pytkdocs.cli.process_config].
    """
    return process_config(json.loads(json_input))


def extract_docstring_parsing_errors(errors: dict, o: Object) -> None:
    """
    Recursion helper.

    Update the `errors` dictionary by side-effect. Recurse on the object's children.

    Arguments:
        errors: The dictionary to update.
        o: The object.
    """
    if hasattr(o, "docstring_errors"):
        errors[o.path] = o.docstring_errors
    for child in o.children:
        extract_docstring_parsing_errors(errors, child)


def extract_errors(obj: Object) -> dict:
    """
    Extract the docstring parsing errors of each object, recursively, into a flat dictionary.

    Arguments:
        obj: An object from `pytkdocs.objects`.

    Returns:
        A flat dictionary. Keys are the objects' names.
    """
    parsing_errors: Dict[str, List[str]] = {}
    extract_docstring_parsing_errors(parsing_errors, obj)
    return parsing_errors


def get_parser() -> argparse.ArgumentParser:
    """Return the program argument parser."""
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-1",
        "--line-by-line",
        action="store_true",
        dest="line_by_line",
        help="Process each line read on stdin, one by one.",
    )
    return parser


def main(args: Optional[Sequence[str]] = None) -> int:
    """
    The main function, which is executed when you type `pytkdocs` or `python -m pytkdocs`.

    Parameters:
        args: The list of arguments.

    Returns:
        An exit code between 0 and 255.
    """
    parser = get_parser()
    parsed_args: argparse.Namespace = parser.parse_args(args)  # type: ignore

    if parsed_args.line_by_line:
        for line in sys.stdin:
            try:
                print(json.dumps(process_json(line)))
            except Exception as error:
                # Don't fail on error. We must handle the next inputs.
                # Instead, print error as JSON.
                print(json.dumps({"error": str(error), "traceback": traceback.format_exc()}))
    else:
        print(json.dumps(process_json(sys.stdin.read())))

    return 0



"""
This module defines function to serialize objects.

These functions simply take objects as parameters and return dictionaries that can be dumped by `json.dumps`.
"""


import inspect
from typing import Optional

from .objects import Object, Source
from .parsers.docstrings import AnnotatedObject, Parameter, Section, annotation_to_string


def serialize_annotated_object(obj: AnnotatedObject) -> dict:
    """
    Serialize an instance of [`AnnotatedObject`][pytkdocs.parsers.docstrings.AnnotatedObject].

    Arguments:
        obj: The object to serialize.

    Returns:
        A JSON-serializable dictionary.
    """
    return dict(description=obj.description, annotation=obj.annotation_string)


def serialize_parameter(parameter: Parameter) -> dict:
    """
    Serialize an instance of [`Parameter`][pytkdocs.parsers.docstrings.Parameter].

    Arguments:
        parameter: The parameter to serialize.

    Returns:
        A JSON-serializable dictionary.
    """
    serialized = serialize_annotated_object(parameter)
    serialized.update(
        dict(
            name=parameter.name,
            kind=str(parameter.kind),
            default=parameter.default_string,
            is_optional=parameter.is_optional,
            is_required=parameter.is_required,
            is_args=parameter.is_args,
            is_kwargs=parameter.is_kwargs,
        )
    )
    return serialized


def serialize_signature_parameter(parameter: inspect.Parameter) -> dict:
    """
    Serialize an instance of `inspect.Parameter`.

    Arguments:
        parameter: The parameter to serialize.

    Returns:
        A JSON-serializable dictionary.
    """
    serialized = dict(kind=str(parameter.kind), name=parameter.name)
    if parameter.annotation is not parameter.empty:
        serialized["annotation"] = annotation_to_string(parameter.annotation)
    if parameter.default is not parameter.empty:
        serialized["default"] = repr(parameter.default)
    return serialized


def serialize_signature(signature: inspect.Signature) -> dict:
    """
    Serialize an instance of `inspect.Signature`.

    Arguments:
        signature: The signature to serialize.

    Returns:
        A JSON-serializable dictionary.
    """
    if signature is None:
        return {}
    serialized: dict = dict(
        parameters=[serialize_signature_parameter(value) for name, value in signature.parameters.items()]
    )
    if signature.return_annotation is not inspect.Signature.empty:
        serialized["return_annotation"] = annotation_to_string(signature.return_annotation)
    return serialized


def serialize_docstring_section(section: Section) -> dict:
    """
    Serialize an instance of `inspect.Signature`.

    Arguments:
        section: The section to serialize.

    Returns:
        A JSON-serializable dictionary.
    """
    serialized = dict(type=section.type)
    if section.type == section.Type.MARKDOWN:
        serialized.update(dict(value="\n".join(section.value)))
    elif section.type == section.Type.RETURN:
        serialized.update(dict(value=serialize_annotated_object(section.value)))
    elif section.type == section.Type.EXCEPTIONS:
        serialized.update(dict(value=[serialize_annotated_object(e) for e in section.value]))
    elif section.type == section.Type.PARAMETERS:
        serialized.update(dict(value=[serialize_parameter(p) for p in section.value]))
    return serialized


def serialize_source(source: Optional[Source]) -> dict:
    """
    Serialize an instance of [`Source`][pytkdocs.objects.Source].

    Arguments:
        source: The source to serialize.

    Returns:
        A JSON-serializable dictionary.
    """
    if source:
        return dict(code=source.code, line_start=source.line_start)
    return {}


def serialize_object(obj: Object) -> dict:
    """
    Serialize an instance of a subclass of [`Object`][pytkdocs.objects.Object].

    Arguments:
        obj: The object to serialize.

    Returns:
        A JSON-serializable dictionary.
    """
    serialized = dict(
        name=obj.name,
        path=obj.path,
        category=obj.category,
        file_path=obj.file_path,
        relative_file_path=obj.relative_file_path,
        properties=sorted(set(obj.properties + obj.name_properties)),
        parent_path=obj.parent_path,
        has_contents=obj.has_contents(),
        docstring=obj.docstring,
        docstring_sections=[serialize_docstring_section(s) for s in obj.docstring_sections],
        source=serialize_source(obj.source),
        children={child.path: serialize_object(child) for child in obj.children},
        attributes=[o.path for o in obj.attributes],
        methods=[o.path for o in obj.methods],
        functions=[o.path for o in obj.functions],
        modules=[o.path for o in obj.modules],
        classes=[o.path for o in obj.classes],
    )
    if hasattr(obj, "type"):
        serialized["type"] = annotation_to_string(obj.type)  # type: ignore
    if hasattr(obj, "signature"):
        serialized["signature"] = serialize_signature(obj.signature)  # type: ignore
    return serialized



"""This module defines functions and classes to parse docstrings into structured data."""

import inspect
import re
from textwrap import dedent
from typing import Any, List, Optional, Pattern, Sequence, Tuple

try:
    from typing import GenericMeta  # python 3.6
except ImportError:
    # in 3.7, GenericMeta doesn't exist but we don't need it
    class GenericMeta(type):  # type: ignore
        pass


empty = inspect.Signature.empty


TITLES_PARAMETERS: Sequence[str] = ("args:", "arguments:", "params:", "parameters:")
"""Titles to match for "parameters" sections."""

TITLES_EXCEPTIONS: Sequence[str] = ("raise:", "raises:", "except:", "exceptions:")
"""Titles to match for "exceptions" sections."""

TITLES_RETURN: Sequence[str] = ("return:", "returns:")
"""Titles to match for "returns" sections."""


RE_OPTIONAL: Pattern = re.compile(r"Union\[(.+), NoneType\]")
"""Regular expression to match optional annotations of the form `Union[T, NoneType]`."""

RE_FORWARD_REF: Pattern = re.compile(r"_?ForwardRef\('([^']+)'\)")
"""Regular expression to match forward-reference annotations of the form `_ForwardRef('T')`."""

RE_GOOGLE_STYLE_ADMONITION: Pattern = re.compile(r"^(?P<indent>\s*)(?P<type>[\w-]+):((?:\s+)(?P<title>.+))?$")
"""Regular expressions to match lines starting admonitions, of the form `TYPE: [TITLE]`."""


class AnnotatedObject:
    """A helper class to store information about an annotated object."""

    def __init__(self, annotation, description):
        self.annotation = annotation
        self.description = description

    def __str__(self):
        return self.annotation_string

    @property
    def annotation_string(self):
        return annotation_to_string(self.annotation)


class Parameter(AnnotatedObject):
    """A helper class to store information about a signature parameter."""

    def __init__(self, name, annotation, description, kind, default=empty):
        super().__init__(annotation, description)
        self.name = name
        self.kind = kind
        self.default = default

    def __str__(self):
        return self.name

    def __repr__(self):
        return f"<Parameter({self.name}, {self.annotation}, {self.description}, {self.kind}, {self.default})>"

    @property
    def is_optional(self):
        return self.default is not empty

    @property
    def is_required(self):
        return not self.is_optional

    @property
    def is_args(self):
        return self.kind is inspect.Parameter.VAR_POSITIONAL

    @property
    def is_kwargs(self):
        return self.kind is inspect.Parameter.VAR_KEYWORD

    @property
    def default_string(self):
        if self.is_kwargs:
            return "{}"
        elif self.is_args:
            return "()"
        elif self.is_required:
            return ""
        return repr(self.default)


class Section:
    """A helper class to store a docstring section."""

    class Type:
        MARKDOWN = "markdown"
        PARAMETERS = "parameters"
        EXCEPTIONS = "exceptions"
        RETURN = "return"

    def __init__(self, section_type, value):
        self.type = section_type
        self.value = value

    def __str__(self):
        return self.type

    def __repr__(self):
        return f"<Section(type={self.type!r})>"


class DocstringParser:
    """
    A class to parse docstrings.

    It is instantiated with an object's path, docstring, signature and return type.

    The `parse` method then returns structured data,
    in the form of a list of [`Section`][pytkdocs.parsers.docstrings.Section]s.
    It also return the list of errors that occurred during parsing.
    """

    def __init__(
        self,
        path: str,
        docstring: str,
        signature: Optional[inspect.Signature] = None,
        return_type: Optional[Any] = empty,
    ) -> None:
        """
        Arguments:
            path: An object's dotted-path, used to improve error messages.
            docstring: An object's docstring: the docstring to parse.
            signature: An object's signature if any.
            return_type: An object's return type if any. Can be a string or a type.
        """
        self.path = path
        self.docstring = docstring or ""
        self.signature = signature
        self.return_type = return_type
        self.parsing_errors: List[str] = []

    def parse(self, admonitions: bool = True) -> List[Section]:
        """
        Parse a docstring.

        Arguments:
            admonitions: Whether to transform "Google-Style admonitions" to "Markdown admonitions"
                by transforming `Type: [Title]` to `!!! type: ["Title"]`.

        Returns:
             A tuple containing the list of parsed sections and the errors that occurred during parsing.
        """
        sections = []
        current_section = []

        in_code_block = False

        lines = self.docstring.split("\n")
        i = 0

        while i < len(lines):
            line_lower = lines[i].lower()
            if in_code_block:
                if line_lower.lstrip(" ").startswith("```"):
                    in_code_block = False
                current_section.append(lines[i])
            elif line_lower in TITLES_PARAMETERS:
                if current_section:
                    if any(current_section):
                        sections.append(Section(Section.Type.MARKDOWN, current_section))
                    current_section = []
                section, i = self.read_parameters_section(lines, i + 1)
                if section:
                    sections.append(section)
            elif line_lower in TITLES_EXCEPTIONS:
                if current_section:
                    if any(current_section):
                        sections.append(Section(Section.Type.MARKDOWN, current_section))
                    current_section = []
                section, i = self.read_exceptions_section(lines, i + 1)
                if section:
                    sections.append(section)
            elif line_lower in TITLES_RETURN:
                if current_section:
                    if any(current_section):
                        sections.append(Section(Section.Type.MARKDOWN, current_section))
                    current_section = []
                section, i = self.read_return_section(lines, i + 1)
                if section:
                    sections.append(section)
            elif line_lower.lstrip(" ").startswith("```"):
                in_code_block = True
                current_section.append(lines[i])
            else:
                if admonitions and not in_code_block and i + 1 < len(lines):
                    match = RE_GOOGLE_STYLE_ADMONITION.match(lines[i])
                    if match:
                        groups = match.groupdict()
                        indent = groups["indent"]
                        if lines[i + 1].startswith(indent + " " * 4):
                            lines[i] = f"{indent}!!! {groups['type'].lower()}"
                            if groups["title"]:
                                lines[i] += f' "{groups["title"]}"'
                current_section.append(lines[i])
            i += 1

        if current_section:
            sections.append(Section(Section.Type.MARKDOWN, current_section))

        return sections

    @staticmethod
    def read_block_items(lines: List[str], start_index: int) -> Tuple[List[str], int]:
        """
        Parse an indented block as a list of items.

        Each item is indented by four spaces. Every line indented with more than five spaces are concatenated
        back into the previous line.

        Arguments:
            lines: The block lines.
            start_index: The line number to start at.

        Returns:
            A tuple containing the list of concatenated lines and the index at which to continue parsing.
        """
        i = start_index
        block: List[str] = []
        prefix = " "
        while i < len(lines) and (lines[i].startswith("    ") or not lines[i].strip()):
            if block and lines[i].startswith("      "):
                block[-1] += prefix + lines[i].lstrip(" ")
                prefix = " "
            elif block and not lines[i].strip():
                block[-1] += "\n\n"
                prefix = ""
            else:
                block.append(lines[i])
            i += 1
        cleaned_up_block = []
        for line in block:
            stripped = line.strip()
            if stripped:
                cleaned_up_block.append(stripped)
        return cleaned_up_block, i - 1

    @staticmethod
    def read_block(lines: List[str], start_index: int) -> Tuple[List[str], int]:
        """
        Parse an indented block.

        Arguments:
            lines: The block lines.
            start_index: The line number to start at.

        Returns:
            A tuple containing the list of lines and the index at which to continue parsing.
        """
        i = start_index
        block = []
        while i < len(lines) and (lines[i].startswith("    ") or not lines[i].strip()):
            block.append(lines[i])
            i += 1
        return block, i - 1

    def read_parameters_section(self, lines: List[str], start_index: int) -> Tuple[Optional[Section], int]:
        """
        Parse a "parameters" section.

        Arguments:
            lines: The parameters block lines.
            start_index: The line number to start at.

        Returns:
            A tuple containing a `Section` (or `None`) and the index at which to continue parsing.
        """
        parameters = []
        type_: Any
        block, i = self.read_block_items(lines, start_index)
        for param_line in block:
            try:
                name_with_type, description = param_line.lstrip(" ").split(":", 1)
            except ValueError:
                self.parsing_errors.append(f"{self.path}: Failed to get 'name: description' pair from '{param_line}'")
                continue

            if " " in name_with_type:
                name, type_ = name_with_type.split(" ", 1)
                type_ = type_.strip("()")
                if type_.endswith(", optional"):
                    type_ = type_[:-10]
            else:
                name = name_with_type
                type_ = empty

            default = empty
            annotation = type_
            kind = None

            try:
                signature_param = self.signature.parameters[name]  # type: ignore
            except (AttributeError, KeyError):
                self.parsing_errors.append(f"{self.path}: No type annotation for parameter '{name}'")
            else:
                if signature_param.annotation is not empty:
                    annotation = signature_param.annotation
                if signature_param.default is not empty:
                    default = signature_param.default
                kind = signature_param.kind

            parameters.append(
                Parameter(
                    name=name, annotation=annotation, description=description.lstrip(" "), default=default, kind=kind,
                )
            )

        if parameters:
            return Section(Section.Type.PARAMETERS, parameters), i

        self.parsing_errors.append(f"{self.path}: Empty parameters section at line {start_index}")
        return None, i

    def read_exceptions_section(self, lines: List[str], start_index: int) -> Tuple[Optional[Section], int]:
        """
        Parse an "exceptions" section.

        Arguments:
            lines: The exceptions block lines.
            start_index: The line number to start at.

        Returns:
            A tuple containing a `Section` (or `None`) and the index at which to continue parsing.
        """
        exceptions = []
        block, i = self.read_block_items(lines, start_index)
        for exception_line in block:
            try:
                annotation, description = exception_line.split(": ", 1)
            except ValueError:
                self.parsing_errors.append(
                    f"{self.path}: Failed to get 'exception: description' pair from '{exception_line}'"
                )
            else:
                exceptions.append(AnnotatedObject(annotation, description.lstrip(" ")))
        if exceptions:
            return Section(Section.Type.EXCEPTIONS, exceptions), i

        self.parsing_errors.append(f"{self.path}: Empty exceptions section at line {start_index}")
        return None, i

    def read_return_section(self, lines: List[str], start_index: int) -> Tuple[Optional[Section], int]:
        """
        Parse an "returns" section.

        Arguments:
            lines: The return block lines.
            start_index: The line number to start at.

        Returns:
            A tuple containing a `Section` (or `None`) and the index at which to continue parsing.
        """
        block, i = self.read_block(lines, start_index)
        if self.signature:
            annotation = self.signature.return_annotation
        else:
            annotation = self.return_type

        if annotation is empty:
            if not block:
                self.parsing_errors.append(f"{self.path}: No return type annotation")
            else:
                try:
                    type_, first_line = block[0].split(":", 1)
                except ValueError:
                    self.parsing_errors.append(f"{self.path}: No type in return description")
                else:
                    annotation = type_.lstrip(" ")
                    block[0] = first_line.lstrip(" ")

        description = dedent("\n".join(block))
        if annotation is empty and not description:
            self.parsing_errors.append(f"{self.path}: Empty return section at line {start_index}")
            return None, i

        return Section(Section.Type.RETURN, AnnotatedObject(annotation, description)), i


def rebuild_optional(matched_group: str) -> str:
    brackets_level = 0
    for char in matched_group:
        if char == "," and brackets_level == 0:
            return f"Union[{matched_group}]"
        elif char == "[":
            brackets_level += 1
        elif char == "]":
            brackets_level -= 1
    return matched_group


def annotation_to_string(annotation: Any):
    if annotation is empty:
        return ""

    if inspect.isclass(annotation) and not isinstance(annotation, GenericMeta):
        s = annotation.__name__
    else:
        s = str(annotation).replace("typing.", "")

    s = RE_FORWARD_REF.sub(lambda match: match.group(1), s)
    s = RE_OPTIONAL.sub(lambda match: f"Optional[{rebuild_optional(match.group(1))}]", s)

    return s


def parse(
    path: str,
    docstring: str,
    signature: Optional[inspect.Signature] = None,
    return_type: Optional[Any] = empty,
    admonitions: bool = True,
):
    parser = DocstringParser(path, docstring, signature, return_type)
    return parser.parse(admonitions), parser.parsing_errors



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['pytkdocs', 'pytkdocs.parsers']

package_data = \
{'': ['*']}

entry_points = \
{'console_scripts': ['pytkdocs = pytkdocs.cli:main']}

setup_kwargs = {
    'name': 'pytkdocs',
    'version': '0.2.1',
    'description': 'Load Python objects documentation.',
    'long_description': '# pytkdocs\n[![pipeline status](https://gitlab.com/pawamoy/pytkdocs/badges/master/pipeline.svg)](https://gitlab.com/pawamoy/pytkdocs/pipelines)\n[![coverage report](https://gitlab.com/pawamoy/pytkdocs/badges/master/coverage.svg)](https://gitlab.com/pawamoy/pytkdocs/commits/master)\n[![documentation](https://img.shields.io/badge/docs-latest-green.svg?style=flat)](https://pawamoy.github.io/pytkdocs)\n[![pypi version](https://img.shields.io/pypi/v/pytkdocs.svg)](https://pypi.org/project/pytkdocs/)\n\nLoad Python objects documentation.\n\n## Requirements\n`pytkdocs` requires Python 3.6 or above.\n\n<details>\n<summary>To install Python 3.6, I recommend using <a href="https://github.com/pyenv/pyenv"><code>pyenv</code></a>.</summary>\n\n```bash\n# install pyenv\ngit clone https://github.com/pyenv/pyenv ~/.pyenv\n\n# setup pyenv (you should also put these three lines in .bashrc or similar)\nexport PATH="${HOME}/.pyenv/bin:${PATH}"\nexport PYENV_ROOT="${HOME}/.pyenv"\neval "$(pyenv init -)"\n\n# install Python 3.6\npyenv install 3.6.8\n\n# make it available globally\npyenv global system 3.6.8\n```\n</details>\n\n## Installation\nWith `pip`:\n```bash\npython3.6 -m pip install pytkdocs\n```\n\nWith [`pipx`](https://github.com/pipxproject/pipx):\n```bash\npython3.6 -m pip install --user pipx\n\npipx install --python python3.6 pytkdocs\n```\n\n## Usage\n\n`pytkdocs` accepts JSON on standard input and writes JSON on standard output.\n\nInput format:\n\n```json\n{\n  "objects": [\n    {\n      "path": "my_module.my_class",\n      "members": true,\n      "filters": [\n        "!^_[^_]"\n      ]\n    }\n  ]\n}\n```\n\n## Configuration\n\nThe configuration options available are:\n\n- `filters`: filters are regular expressions that allow to select or un-select objects based on their name.\n  They are applied recursively (on every child of every object).\n  If the expression starts with an exclamation mark,\n  it will filter out objects matching it (the exclamation mark is removed before evaluation).\n  If not, objects matching it are selected.\n  Every regular expression is performed against every name.\n  It allows fine-grained filtering. Example:\n\n    - `!^_`: filter out every object whose name starts with `_` (private/protected)\n    - `^__`: but still select those who start with two `_` (class-private)\n    - `!^__.*__$`: except those who also end with two `_` (specials)\n  \n- `members`: this option allows to explicitly select the members of the top-object.\n  If `True`, select every members that passes filters. If `False`, select nothing.\n  If it\'s a list of names, select only those members, and apply filters on their children only.\n',
    'author': 'Timothée Mazzucotelli',
    'author_email': 'pawamoy@pm.me',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/pawamoy/pytkdocs',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



"""
This module is responsible for loading the documentation from Python objects.

It uses [`inspect`](https://docs.python.org/3/library/inspect.html) for introspecting objects,
iterating over their members, etc.
"""

import importlib
import inspect
import pkgutil
import re
import textwrap
from functools import lru_cache
from pathlib import Path
from typing import Any, List, Optional, Set, Union

from .objects import Attribute, Class, Function, Method, Module, Object, Source
from .parsers.attributes import get_attributes
from .properties import RE_SPECIAL


class ObjectNode:
    """
    Helper class to represent an object tree.

    It's not really a tree but more a backward-linked list:
    each node has a reference to its parent, but not to its child (for simplicity purposes and to avoid bugs).

    Each node stores an object, its name, and a reference to its parent node.
    """

    def __init__(self, obj: Any, name: str, parent: Optional["ObjectNode"] = None) -> None:
        self.obj: Any = obj
        """The actual Python object."""

        self.name: str = name
        """The Python object's name."""

        self.parent: Optional[ObjectNode] = parent
        """The parent node."""

    @property
    def dotted_path(self) -> str:
        """The Python dotted path of the object."""
        parts = [self.name]
        current = self.parent
        while current:
            parts.append(current.name)
            current = current.parent
        return ".".join(reversed(parts))

    @property
    def file_path(self) -> str:
        """The object's module file path."""
        return inspect.getabsfile(self.root.obj)

    @property
    def root(self) -> "ObjectNode":
        """The root of the tree."""
        if self.parent is not None:
            return self.parent.root
        return self

    def is_module(self) -> bool:
        """Is this node's object a module?"""
        return inspect.ismodule(self.obj)

    def is_class(self) -> bool:
        """Is this node's object a class?"""
        return inspect.isclass(self.obj)

    def is_function(self) -> bool:
        """Is this node's object a function?"""
        return inspect.isfunction(self.obj)

    def is_property(self) -> bool:
        """Is this node's object a property?"""
        return isinstance(self.obj, property)

    def parent_is_class(self) -> bool:
        """Is the object of this node's parent a class?"""
        return bool(self.parent and self.parent.is_class())

    def is_method(self) -> bool:
        """Is this node's object a method?"""
        return self.parent_is_class() and isinstance(self.obj, type(lambda: 0))

    def is_staticmethod(self) -> bool:
        """Is this node's object a staticmethod?"""
        if not self.parent:
            return False
        return self.parent_is_class() and isinstance(self.parent.obj.__dict__.get(self.name, None), staticmethod)

    def is_classmethod(self) -> bool:
        """Is this node's object a classmethod?"""
        if not self.parent:
            return False
        return self.parent_is_class() and isinstance(self.parent.obj.__dict__.get(self.name, None), classmethod)


def get_object_tree(path: str) -> ObjectNode:
    """
    Transform a path into an actual Python object.

    The path can be arbitrary long. You can pass the path to a package,
    a module, a class, a function or a global variable, as deep as you
    want, as long as the deepest module is importable through
    `importlib.import_module` and each object is obtainable through
    the `getattr` method. It is not possible to load local objects.

    Args:
        path: the dot-separated path of the object.

    Raises:
        ValueError: when the path is not valid (evaluates to `False`).
        ImportError: when the object or its parent module could not be imported.

    Returns:
        The leaf node representing the object and its parents.
    """
    if not path:
        raise ValueError(f"path must be a valid Python path, not {path}")

    # We will try to import the longest dotted-path first.
    # If it fails, we remove the right-most part and put it in a list of "objects", used later.
    # We loop until we find the deepest importable submodule.
    obj_parent_modules = path.split(".")
    objects: List[str] = []

    while True:
        parent_module_path = ".".join(obj_parent_modules)
        try:
            parent_module = importlib.import_module(parent_module_path)
        except ImportError:
            if len(obj_parent_modules) == 1:
                raise ImportError("No module named '%s'" % obj_parent_modules[0])
            objects.insert(0, obj_parent_modules.pop(-1))
        else:
            break

    # We now have the module containing the desired object.
    # We will build the object tree by iterating over the previously stored objects names
    # and trying to get them as attributes.
    current_node = ObjectNode(parent_module, parent_module.__name__)
    for obj_name in objects:
        obj = getattr(current_node.obj, obj_name)
        current_node.child = ObjectNode(obj, obj_name, parent=current_node)  # type: ignore
        current_node = current_node.child  # type: ignore

    leaf = current_node

    # We now try to get the "real" parent module, not the one the object was imported into.
    # This is important if we want to be able to retrieve the docstring of an attribute for example.
    # Once we find an object for which we could get the module, we stop trying to get the module.
    # Once we reach the node before the root, we apply the module if found, and break.
    real_module = None
    while current_node.parent is not None:
        if real_module is None:
            real_module = inspect.getmodule(current_node.obj)
        if inspect.ismodule(current_node.parent.obj):
            if real_module is not None and real_module is not current_node.parent.obj:
                current_node.parent = ObjectNode(real_module, real_module.__name__)
            break
        current_node = current_node.parent

    return leaf


class Loader:
    """
    This class contains the object documentation loading mechanisms.

    Any error that occurred during collection of the objects and their documentation is stored in the `errors` list.
    """

    def __init__(self, filters: Optional[List[str]] = None):
        """
        Arguments:
            filters: A list of regular expressions to fine-grain select members. It is applied recursively.
        """
        if not filters:
            filters = []

        self.filters = [(f, re.compile(f.lstrip("!"))) for f in filters]
        self.errors: List[str] = []

    def get_object_documentation(self, dotted_path: str, members: Optional[Union[Set[str], bool]] = None) -> Object:
        """
        Get the documentation for an object and its children.

        Arguments:
            dotted_path: The Python dotted path to the desired object.
            members: `True` to select members and filter them, `False` to select no members,
                or a list of names to explicitly select the members with these names.
                It is applied only on the root object.

        Return:
            The documented object.
        """

        if members is True:
            members = set()

        root_object: Object
        leaf = get_object_tree(dotted_path)

        attributes = get_attributes(leaf.root.obj)

        if leaf.is_module():
            root_object = self.get_module_documentation(leaf, members)
        elif leaf.is_class():
            root_object = self.get_class_documentation(leaf, members)
        elif leaf.is_staticmethod():
            root_object = self.get_staticmethod_documentation(leaf)
        elif leaf.is_classmethod():
            root_object = self.get_classmethod_documentation(leaf)
        elif leaf.is_method():
            root_object = self.get_regular_method_documentation(leaf)
        elif leaf.is_function():
            root_object = self.get_function_documentation(leaf)
        elif leaf.is_property():
            root_object = self.get_property_documentation(leaf)
        else:
            for attribute in attributes:
                if attribute.path == dotted_path:
                    return attribute
            raise ValueError(f"{dotted_path}: {type(leaf.obj)} not yet supported")

        if members is not False:
            filtered = []
            for attribute in attributes:
                if attribute.parent_path == root_object.path:
                    if self.select(attribute.name, members):  # type: ignore
                        filtered.append(attribute)
                elif self.select(attribute.name, set()):
                    filtered.append(attribute)
            root_object.dispatch_attributes(filtered)

        root_object.parse_all_docstring()

        return root_object

    def get_module_documentation(self, node: ObjectNode, members=None) -> Module:
        """
        Get the documentation for a module and its children.

        Arguments:
            node: The node representing the module and its parents.
            members: Explicit members to select.

        Return:
            The documented module object.
        """
        module = node.obj
        path = node.dotted_path
        name = path.split(".")[-1]
        source: Optional[Source]

        try:
            source = Source(inspect.getsource(module), 1)
        except OSError as error:
            try:
                with Path(node.file_path).open() as fd:
                    contents = fd.readlines()
                    if contents:
                        source = Source(contents, 1)
                    else:
                        source = None
            except OSError:
                self.errors.append(f"Couldn't read source for '{path}': {error}")
                source = None

        root_object = Module(
            name=name, path=path, file_path=node.file_path, docstring=inspect.getdoc(module) or "", source=source,
        )

        if members is False:
            return root_object

        members = members or set()

        for member_name, member in inspect.getmembers(module):
            if self.select(member_name, members):  # type: ignore
                child_node = ObjectNode(member, member_name, parent=node)
                if child_node.is_class() and child_node.root.obj is inspect.getmodule(member):
                    root_object.add_child(self.get_class_documentation(child_node))
                elif child_node.is_function() and child_node.root.obj is inspect.getmodule(member):
                    root_object.add_child(self.get_function_documentation(child_node))

        try:
            package_path = module.__path__
        except AttributeError:
            pass
        else:
            for _, modname, _ in pkgutil.iter_modules(package_path):
                if self.select(modname, members):
                    leaf = get_object_tree(f"{path}.{modname}")
                    root_object.add_child(self.get_module_documentation(leaf))

        return root_object

    def get_class_documentation(self, node: ObjectNode, members=None) -> Class:
        """
        Get the documentation for a class and its children.

        Arguments:
            node: The node representing the class and its parents.
            members: Explicit members to select.

        Return:
            The documented class object.
        """
        class_ = node.obj
        docstring = textwrap.dedent(class_.__doc__ or "")
        root_object = Class(name=node.name, path=node.dotted_path, file_path=node.file_path, docstring=docstring)

        if members is False:
            return root_object

        members = members or set()

        for member_name, member in class_.__dict__.items():
            if member is type or member is object:
                continue

            if not self.select(member_name, members):  # type: ignore
                continue

            child_node = ObjectNode(getattr(class_, member_name), member_name, parent=node)
            if child_node.is_class():
                root_object.add_child(self.get_class_documentation(child_node))
            elif child_node.is_classmethod():
                root_object.add_child(self.get_classmethod_documentation(child_node))
            elif child_node.is_staticmethod():
                root_object.add_child(self.get_staticmethod_documentation(child_node))
            elif child_node.is_method():
                root_object.add_child(self.get_regular_method_documentation(child_node))
            elif child_node.is_property():
                root_object.add_child(self.get_property_documentation(child_node))

        return root_object

    def get_function_documentation(self, node: ObjectNode) -> Function:
        """
        Get the documentation for a function.

        Arguments:
            node: The node representing the function and its parents.

        Return:
            The documented function object.
        """
        function = node.obj
        path = node.dotted_path
        source: Optional[Source]
        signature: Optional[inspect.Signature]

        try:
            signature = inspect.signature(function)
        except TypeError as error:
            self.errors.append(f"Couldn't get signature for '{path}': {error}")
            signature = None

        try:
            source = Source(*inspect.getsourcelines(function))
        except OSError as error:
            self.errors.append(f"Couldn't read source for '{path}': {error}")
            source = None

        return Function(
            name=node.name,
            path=node.dotted_path,
            file_path=node.file_path,
            docstring=inspect.getdoc(function),
            signature=signature,
            source=source,
        )

    def get_property_documentation(self, node: ObjectNode) -> Attribute:
        """
        Get the documentation for an attribute.

        Arguments:
            node: The node representing the attribute and its parents.

        Return:
            The documented attribute object.
        """
        prop = node.obj
        path = node.dotted_path
        properties = ["property", "readonly" if prop.fset is None else "writable"]
        source: Optional[Source]

        try:
            signature = inspect.signature(prop.fget)
        except (TypeError, ValueError) as error:
            self.errors.append(f"Couldn't get signature for '{path}': {error}")
            attr_type = None
        else:
            attr_type = signature.return_annotation

        try:
            source = Source(*inspect.getsourcelines(prop.fget))
        except (OSError, TypeError) as error:
            self.errors.append(f"Couldn't get source for '{path}': {error}")
            source = None

        return Attribute(
            name=node.name,
            path=path,
            file_path=node.file_path,
            docstring=inspect.getdoc(prop.fget),
            attr_type=attr_type,
            properties=properties,
            source=source,
        )

    def get_classmethod_documentation(self, node: ObjectNode) -> Method:
        """
        Get the documentation for a class-method.

        Arguments:
            node: The node representing the class-method and its parents.

        Return:
            The documented method object.
        """
        return self.get_method_documentation(node, ["classmethod"])

    def get_staticmethod_documentation(self, node: ObjectNode) -> Method:
        """
        Get the documentation for a static-method.

        Arguments:
            node: The node representing the static-method and its parents.

        Return:
            The documented method object.
        """
        return self.get_method_documentation(node, ["staticmethod"])

    def get_regular_method_documentation(self, node: ObjectNode) -> Method:
        """
        Get the documentation for a regular method (not class- nor static-method).

        We do extra processing in this method to discard docstrings of `__init__` methods
        that were inherited from parent classes.

        Arguments:
            node: The node representing the method and its parents.

        Return:
            The documented method object.
        """
        method = self.get_method_documentation(node)
        if node.parent:
            class_ = node.parent.obj
            if RE_SPECIAL.match(node.name):
                docstring = method.docstring
                parent_classes = class_.__mro__[1:]
                for parent_class in parent_classes:
                    try:
                        parent_method = getattr(parent_class, node.name)
                    except AttributeError:
                        continue
                    else:
                        if docstring == inspect.getdoc(parent_method):
                            method.docstring = ""
                        break
        return method

    def get_method_documentation(self, node: ObjectNode, properties: Optional[List[str]] = None) -> Method:
        """
        Get the documentation for a method.

        Arguments:
            node: The node representing the method and its parents.
            properties: A list of properties to apply to the method.

        Return:
            The documented method object.
        """
        method = node.obj
        path = node.dotted_path
        source: Optional[Source]

        try:
            source = Source(*inspect.getsourcelines(method))
        except OSError as error:
            self.errors.append(f"Couldn't read source for '{path}': {error}")
            source = None
        except TypeError:
            source = None

        return Method(
            name=node.name,
            path=path,
            file_path=node.file_path,
            docstring=inspect.getdoc(method),
            signature=inspect.signature(method),
            properties=properties or [],
            source=source,
        )

    def select(self, name: str, names: Set[str]) -> bool:
        """
        Tells whether we should select an object or not, given its name.

        If the set of names is not empty, we check against it, otherwise we check against filters.

        Arguments:
            name: The name of the object to select or not.
            names: An explicit list of names to select.

        Returns:
            Yes or no.
        """
        if names:
            return name in names
        return not self.filter_name_out(name)

    @lru_cache(maxsize=None)
    def filter_name_out(self, name: str) -> bool:
        if not self.filters:
            return False
        keep = True
        for f, regex in self.filters:
            is_matching = bool(regex.search(name))
            if is_matching:
                if str(f).startswith("!"):
                    is_matching = not is_matching
                keep = is_matching
        return not keep



import ast
import inspect
from functools import lru_cache
from textwrap import dedent
from types import ModuleType
from typing import Any, Iterable, List, Optional, Union

from ..objects import Attribute

RECURSIVE_NODES = (ast.If, ast.IfExp, ast.Try, ast.With, ast.ExceptHandler)


def node_is_docstring(node: ast.AST) -> bool:
    return isinstance(node, ast.Expr) and isinstance(node.value, ast.Str)


def node_to_docstring(node: Union[ast.Expr, ast.Str]) -> str:
    return node.value.s  # type: ignore


def node_is_assignment(node: ast.AST) -> bool:
    return isinstance(node, ast.Assign)


def node_is_annotated_assignment(node: ast.AST) -> bool:
    return isinstance(node, ast.AnnAssign)


def node_to_names(node: ast.Assign) -> dict:
    names = []
    for target in node.targets:
        if isinstance(target, ast.Attribute):
            names.append(target.attr)
        elif isinstance(target, ast.Name):
            names.append(target.id)
    return {"names": names, "lineno": node.lineno, "type": inspect.Signature.empty}


def node_to_annotated_names(node: ast.AnnAssign) -> dict:
    try:
        name = node.target.id  # type: ignore
    except AttributeError:
        name = node.target.attr  # type: ignore
    lineno = node.lineno
    return {"names": [name], "lineno": lineno, "type": node_to_annotation(node)}


def node_to_annotation(node) -> Union[str, object]:
    if isinstance(node, ast.AnnAssign):
        try:
            annotation = node.annotation.id  # type: ignore
        except AttributeError:
            if isinstance(node.annotation, ast.Str):
                annotation = node.annotation.s
            else:
                annotation = node.annotation.value.id  # type: ignore
        if hasattr(node.annotation, "slice"):
            annotation += f"[{node_to_annotation(node.annotation.slice.value)}]"  # type: ignore
        return annotation
    elif isinstance(node, ast.Subscript):
        return f"{node.value.id}[{node_to_annotation(node.slice.value)}]"  # type: ignore
    elif isinstance(node, ast.Tuple):
        annotations = [node_to_annotation(n) for n in node.elts]
        return ", ".join(a for a in annotations if a is not inspect.Signature.empty)  # type: ignore
    elif isinstance(node, ast.Name):
        return node.id
    return inspect.Signature.empty


def get_attribute_info(node1, node2):
    if node_is_docstring(node2):
        info = {"docstring": node_to_docstring(node2)}
        if node_is_assignment(node1):
            info.update(node_to_names(node1))
            return info
        elif node_is_annotated_assignment(node1):
            info.update(node_to_annotated_names(node1))
            return info
    raise ValueError(f"nodes must be assignment and docstring, not '{node1}' and '{node2}'")


@lru_cache(maxsize=None)
def get_attributes(module: ModuleType) -> List[Attribute]:
    file_path = module.__file__
    with open(file_path) as stream:
        code = stream.read()
    initial_ast_body = ast.parse(code).body
    return _get_attributes(initial_ast_body, name_prefix=module.__name__, file_path=file_path)


def _get_attributes(
    ast_body: Iterable, name_prefix: str, file_path: str, properties: Optional[List[str]] = None
) -> List[Attribute]:
    if not properties:
        properties = []

    documented_attributes = []
    previous_node = None
    body: Any

    for node in ast_body:
        try:
            attr_info = get_attribute_info(previous_node, node)
        except ValueError:
            if isinstance(node, RECURSIVE_NODES):
                documented_attributes.extend(_get_attributes(node.body, name_prefix, file_path, properties))  # type: ignore
                if isinstance(node, ast.Try):
                    for body in [node.handlers, node.orelse, node.finalbody]:
                        documented_attributes.extend(_get_attributes(body, name_prefix, file_path, properties))
                elif isinstance(node, ast.If):
                    documented_attributes.extend(_get_attributes(node.orelse, name_prefix, file_path, properties))

            elif isinstance(node, ast.FunctionDef) and node.name == "__init__":
                documented_attributes.extend(_get_attributes(node.body, name_prefix, file_path))

            elif isinstance(node, ast.ClassDef):
                documented_attributes.extend(
                    _get_attributes(node.body, f"{name_prefix}.{node.name}", file_path, properties=["class-attribute"])
                )
        else:
            for name in attr_info["names"]:
                documented_attributes.append(
                    Attribute(
                        name=name,
                        path=f"{name_prefix}.{name}",
                        file_path=file_path,
                        docstring=dedent(attr_info["docstring"]),
                        properties=properties,
                        attr_type=attr_info["type"],
                    )
                )
        previous_node = node

    return documented_attributes
